﻿using System;
using System.Collections.Generic;
using System.Management;
using System.Runtime.InteropServices;
using System.Text;

namespace DojoNorthSoftware.HomeServer.HomeServerSMARTService.Enumerator
{
    public class InteropSmart
    {
        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetInternalDriveSmartAttributes(int physicalDriveId, byte target, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetInternalDriveSmartThresholds(int physicalDriveId, byte target, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetUsbDriveSmartAttributes(int physicalDriveId, byte target, COMMAND_TYPE cmdType, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetUsbDriveSmartThresholds(int physicalDriveId, byte target, COMMAND_TYPE cmdType, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetSiIDriveSmartAttributes(int physicalDriveId, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetScsiDriveSmartAttributes(int scsiPort, int scsiTargetId, IntPtr asi, IntPtr smart);

        [DllImport("HomeServerSMART.Utility.dll", SetLastError = true)]
        static extern SMART_QUERY_RESULT GetScsiDriveSmartThresholds(int scsiPort, int scsiTargetId, IntPtr asi, IntPtr smart);


        /// <summary>
        /// 
        /// Gets the SMART attributes and thresholds (SMART data) from "internal" disks -- these basically being anything that
        /// isn't SCSI, FireWire or USB.  We'll regard eSATA as "internal."
        /// </summary>
        /// <param name="physicalDrivePath">The physical drive string from WMI, which is of the format \\.\PHYSICALDRIVEx</param>
        /// <param name="target">The target ID of the disk. Probably will be 0xA0 (160) -- use GlobalConstants.DEFAULT_TARGET_ID.</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetInternalSmartData(String physicalDrivePath, byte target, out byte[] smartData, out byte[] smartThreshold,
            out bool failurePredicted)
        {
            return GetInternalSmartData(physicalDrivePath, GlobalConstants.DEFAULT_TARGET_ID, false, null, out smartData,
                out smartThreshold, out failurePredicted, false);
        }

        /// <summary>
        /// Gets the SMART attributes and thresholds (SMART data) from "internal" disks -- these basically being anything that
        /// isn't SCSI, FireWire or USB.  We'll regard eSATA as "internal."
        /// </summary>
        /// <param name="physicalDrivePath">The physical drive string from WMI, which is of the format \\.\PHYSICALDRIVEx</param>
        /// <param name="target">The target ID of the disk. Probably will be 0xA0 (160) -- use GlobalConstants.DEFAULT_TARGET_ID.</param>
        /// <param name="isDiskSiI">Flag that indicates if the disk is attached to an SiI controller.</param>
        /// <param name="pnpDeviceId">PNP Device ID of the disk drive.</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <param name="alternateSiIMethod">Set to true to use an alternate method to get SiI SMART data (no WMI; all thresholds will be zero).
        /// This parameter has no effect if isDiskSiI is false.</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetInternalSmartData(String physicalDrivePath, byte target, bool isDiskSiI, String pnpDeviceId,
            out byte[] smartData, out byte[] smartThreshold, out bool failurePredicted, bool alternateSiIMethod)
        {
            bool ok = false;
            smartData = null;
            smartThreshold = null;
            // TODO: Get the below from WMI!
            failurePredicted = false;

            // Physical drive string is (without quotes) "\\.\PHYSICALDRIVEx" so we want the index.
            // Index is as follows =====================> 012345678901234567 (17)

            int physicalDriveId = -1;

            try
            {
                String driveId = physicalDrivePath.Substring(17);
                physicalDriveId = Int32.Parse(driveId);
            }
            catch
            {
                return false;
            }

            if (isDiskSiI)
            {
                // The disk is attached to a Silicon Image SiI controller so we use the interop calls for the attributes, and WMI
                // for the thresholds.
                if (GetSiliconImageSmartAttributes(physicalDriveId, out smartData))
                {
                    // Got the SiI attributes, now get the thresholds.
                    if (alternateSiIMethod)
                    {
                        smartThreshold = new byte[512];
                        for (int i = 0; i < 512; i++)
                        {
                            smartThreshold[i] = 0;
                        }
                    }
                    else
                    {
                        ok = GetSiliconImageSmartThresholds(pnpDeviceId, out smartThreshold);
                    }
                }
            }
            else
            {
                // The disk is not attached to an SiI controller so use the standard interop calls to get the attributes and thresholds.
                if (GetInternalSmartAttributes(physicalDriveId, target, out smartData))
                {
                    // Got the SiI attributes, now get the thresholds.
                    ok = GetInternalSmartThresholds(physicalDriveId, target, out smartThreshold);
                }
            }

            return ok;
        }

        /// <summary>
        /// Gets the SMART attributes and thresholds (SMART data) from USB disks. If no COMMAND_TYPE is specified, the default of
        /// COMMAND_TYPE.CMD_TYPE_SAT is used.
        /// </summary>
        /// <param name="physicalDrivePath">The physical drive string from WMI, which is of the format \\.\PHYSICALDRIVEx</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetUsbSmartData(String physicalDrivePath, out byte[] smartData, out byte[] smartThreshold, out bool failurePredicted)
        {
            return GetUsbSmartData(physicalDrivePath, COMMAND_TYPE.CMD_TYPE_SAT, out smartData, out smartThreshold, out failurePredicted);
        }

        /// <summary>
        /// Gets the SMART attributes and thresholds (SMART data) from USB disks. If no COMMAND_TYPE is specified, the default of
        /// COMMAND_TYPE.CMD_TYPE_SAT is used.
        /// </summary>
        /// <param name="physicalDrivePath">The physical drive string from WMI, which is of the format \\.\PHYSICALDRIVEx</param>
        /// <param name="diskCommandType">The COMMAND_TYPE that specifies what type of USB bridge chip such as Sunplus, JMicron, IO Data, etc. If
        /// you're not sure specify COMMAND_TYPE.CMD_TYPE_SAT. Note that only SAT, SUNPLUS, LOGITEC, IO_DATA, JMICRON and CYPRESS should be used
        /// with USB disks. All others will fail with SMART_QUERY_EXCEPTION being returned.</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetUsbSmartData(String physicalDrivePath, COMMAND_TYPE diskCommandType, out byte[] smartData, out byte[] smartThreshold,
            out bool failurePredicted)
        {
            bool ok = false;
            smartData = null;
            smartThreshold = null;
            // TODO: get below from WMI!
            failurePredicted = false;

            // Physical drive string is (without quotes) "\\.\PHYSICALDRIVEx" so we want the index.
            // Index is as follows =====================> 012345678901234567 (17)

            int physicalDriveId = -1;

            try
            {
                String driveId = physicalDrivePath.Substring(17);
                physicalDriveId = Int32.Parse(driveId);
            }
            catch
            {
                return false;
            }

            if (GetUsbSmartAttributes(physicalDriveId, GlobalConstants.DEFAULT_TARGET_ID, diskCommandType, out smartData))
            {
                ok = GetUsbSmartThresholds(physicalDriveId, GlobalConstants.DEFAULT_TARGET_ID, diskCommandType, out smartThreshold);
            }

            return ok;
        }

        /// <summary>
        /// Gets the SMART attributes and thresholds (SMART data) from SCSI disks.
        /// </summary>
        /// <param name="scsiPort">SCSI port number.</param>
        /// <param name="scsiTarget">SCSI target ID.</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetScsiSmartData(int scsiPort, int scsiTarget, out byte[] smartData, out byte[] smartThreshold, out bool failurePredicted)
        {
            bool ok = false;
            smartData = null;
            smartThreshold = null;
            // TODO: get below from WMI!
            failurePredicted = false;

            if( GetScsiSmartAttributes(scsiPort, scsiTarget, out smartData ))
            {
                ok = GetScsiSmartThresholds(scsiPort, scsiTarget, out smartThreshold);
            }

            return ok;
        }

        /// <summary>
        /// Gets the SMART attributes and thresholds (SMART data) using WMI.
        /// </summary>
        /// <param name="physicalDrivePath">The physical drive string from WMI, which is of the format \\.\PHYSICALDRIVEx</param>
        /// <param name="smartData">512-byte array with the SMART attribute data.</param>
        /// <param name="smartThreshold">512-byte array with the SMART threshold data.</param>
        /// <param name="failurePredicted">Flag that indicates whether or not WMI is showing "Failure Predicted."</param>
        /// <returns>true if we successfully got the SMART data; false otherwise.</returns>
        public static bool GetWmiSmartData(String physicalDrivePath, out byte[] smartData, out byte[] smartThreshold, out bool failurePredicted)
        {
            bool ok = false;
            smartData = null;
            smartThreshold = null;
            // TODO: get below from WMI!
            failurePredicted = false;

            if (GetWmiSmartAttributes(physicalDrivePath, out smartData))
            {
                ok = GetWmiSmartThresholds(physicalDrivePath, out smartThreshold);
            }

            return ok;
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART attributes from a USB disk.
        /// </summary>
        /// <param name="physicalDriveId">The physical drive number.</param>
        /// <param name="target">The target ID of the disk (probably will be 0xA0).</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART attributes; false otherwise.</returns>
        private static bool GetUsbSmartAttributes(int physicalDriveId, byte target, COMMAND_TYPE diskCommandType, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                SMART_DATA_UCHAR data = new SMART_DATA_UCHAR();
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA_UCHAR));
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetUsbDriveSmartAttributes(physicalDriveId, target, diskCommandType, pointer);

                SMART_DATA_UCHAR smartData = (SMART_DATA_UCHAR)Marshal.PtrToStructure(pointer, typeof(SMART_DATA_UCHAR));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART thresholds from a USB disk.
        /// </summary>
        /// <param name="physicalDriveId">The physical drive number.</param>
        /// <param name="target">The target ID of the disk (probably will be 0xA0).</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART thresholds; false otherwise.</returns>
        private static bool GetUsbSmartThresholds(int physicalDriveId, byte target, COMMAND_TYPE diskCommandType, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                SMART_DATA_UCHAR data = new SMART_DATA_UCHAR();
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA_UCHAR));
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetUsbDriveSmartThresholds(physicalDriveId, target, diskCommandType, pointer);

                SMART_DATA_UCHAR smartData = (SMART_DATA_UCHAR)Marshal.PtrToStructure(pointer, typeof(SMART_DATA_UCHAR));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART attributes from an internal IDE/SATA/eSATA disk.
        /// </summary>
        /// <param name="physicalDriveId">The physical drive number.</param>
        /// <param name="target">The target ID of the disk (probably will be 0xA0).</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART attributes; false otherwise.</returns>
        private static bool GetInternalSmartAttributes(int physicalDriveId, byte target, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                SMART_DATA data = new SMART_DATA();
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA));
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetInternalDriveSmartAttributes(physicalDriveId, target, pointer);

                SMART_DATA smartData = (SMART_DATA)Marshal.PtrToStructure(pointer, typeof(SMART_DATA));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART thresholds from an internal IDE/SATA/eSATA disk.
        /// </summary>
        /// <param name="physicalDriveId">The physical drive number.</param>
        /// <param name="target">The target ID of the disk (probably will be 0xA0).</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART thresholds; false otherwise.</returns>
        private static bool GetInternalSmartThresholds(int physicalDriveId, byte target, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                SMART_DATA data = new SMART_DATA();
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA));
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetInternalDriveSmartThresholds(physicalDriveId, target, pointer);

                SMART_DATA smartData = (SMART_DATA)Marshal.PtrToStructure(pointer, typeof(SMART_DATA));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Gets the SMART attributes for a disk attached to a Silicon Image SiI controller.
        /// </summary>
        /// <param name="physicalDriveId">The physical device ID.</param>
        /// <param name="dataBuffer">output byte array containing SMART data.</param>
        /// <returns>true if attributes successfully fetched; false otherwise.</returns>
        private static bool GetSiliconImageSmartAttributes(int physicalDriveId, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                SMART_DATA_SII data = new SMART_DATA_SII();
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA_SII));
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetSiIDriveSmartAttributes(physicalDriveId, pointer);

                SMART_DATA_SII smartData = (SMART_DATA_SII)Marshal.PtrToStructure(pointer, typeof(SMART_DATA_SII));
                dataBuffer = smartData.VendorSpecific;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Gets the SMART thresholds for a disk attached to a Silicon Image SiI controller.  This
        /// needs to be done through WMI.
        /// </summary>
        /// <param name="pnpDeviceId">PNP device ID of the disk to evaluate.</param>
        /// <param name="smartThreshold">output byte array containing SMART thresholds.</param>
        /// <returns>true if thresholds successfully fetched; false otherwise.</returns>
        private static bool GetSiliconImageSmartThresholds(String pnpDeviceId, out byte[] smartThreshold)
        {
            smartThreshold = null;

            try
            {
                ManagementObjectSearcher WMISearch = new ManagementObjectSearcher();
                //MSStorageDriver is in \root\wmi namespace.
                WMISearch.Scope = new ManagementScope(@"\root\wmi");

                // Must change \\ to \\\\ or WMI call will die horribly.
                String testDisk = pnpDeviceId.Replace("\\", "\\\\");

                // WMI is slow so rather than querying all the MSStorageDriver objects, query for the specific disk.
                WMISearch.Query = new ObjectQuery("select VendorSpecific from MSStorageDriver_FailurePredictThresholds where InstanceName=\"" + testDisk + "_0\"");
                ManagementObjectCollection FailDataSet = WMISearch.Get();
                foreach (ManagementObject attribute in FailDataSet)
                {
                    // If nothing is returned the foreach will have nothing to iterate.
                    smartThreshold = (Byte[])attribute.Properties["VendorSpecific"].Value;
                    return true;
                }
            }
            catch (Exception)
            {
                smartThreshold = null;
                return false;
            }
            return false;
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART attributes from a SCSI disk.
        /// </summary>
        /// <param name="scsiPort">The SCSI port number.</param>
        /// <param name="scsiTarget">The SCSI target number.</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART attributes; false otherwise.</returns>
        private static bool GetScsiSmartAttributes(int scsiPort, int scsiTarget, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                ATA_SMART_INFO asi = new ATA_SMART_INFO();
                SMART_DATA data = new SMART_DATA();
                uint asiSize = (uint)Marshal.SizeOf(typeof(ATA_SMART_INFO));
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA));
                IntPtr asiPtr = Marshal.AllocHGlobal((int)asiSize);
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(asi, asiPtr, true);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetScsiDriveSmartAttributes(scsiPort, scsiTarget, asiPtr, pointer);

                SMART_DATA smartData = (SMART_DATA)Marshal.PtrToStructure(pointer, typeof(SMART_DATA));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Performs the P/Invoke call to get SMART thresholds from a SCSI disk.
        /// </summary>
        /// <param name="scsiPort">The SCSI port number.</param>
        /// <param name="scsiTarget">The SCSI target number.</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the P/Invoke call.</param>
        /// <returns>true if we got the SMART thresholds; false otherwise.</returns>
        private static bool GetScsiSmartThresholds(int scsiPort, int scsiTarget, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;

            try
            {
                ATA_SMART_INFO asi = new ATA_SMART_INFO();
                SMART_DATA data = new SMART_DATA();
                uint asiSize = (uint)Marshal.SizeOf(typeof(ATA_SMART_INFO));
                uint dataSize = (uint)Marshal.SizeOf(typeof(SMART_DATA));
                IntPtr asiPtr = Marshal.AllocHGlobal((int)asiSize);
                IntPtr pointer = Marshal.AllocHGlobal((int)dataSize);
                Marshal.StructureToPtr(asi, asiPtr, true);
                Marshal.StructureToPtr(data, pointer, true);
                ok = GetScsiDriveSmartThresholds(scsiPort, scsiTarget, asiPtr, pointer);

                SMART_DATA smartData = (SMART_DATA)Marshal.PtrToStructure(pointer, typeof(SMART_DATA));
                dataBuffer = smartData.Data;
            }
            catch
            {
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_IO_CONTROL_FALSE;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Gets the SMART attributes for the specified disk from WMI, if WMI has that data available.
        /// </summary>
        /// <param name="physicalDrivePath">Path to the physical drive.</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the WMI call.</param>
        /// <returns>true if the SMART data was successfully retrieved from WMI; false otherwise.</returns>
        private static bool GetWmiSmartAttributes(String physicalDrivePath, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;
            dataBuffer = null;

            try
            {
                ManagementObjectSearcher WMISearch = new ManagementObjectSearcher();
                WMISearch.Scope = new ManagementScope(@"\root\wmi");
                // Need to convert "\\" to "\\\\" or the WMI call will fail.
                String testDisk = physicalDrivePath.Replace("\\", "\\\\");

                // Not sure why but in WMI the instance name field has a _0 tacked on the end of it.
                WMISearch.Query = new ObjectQuery("select VendorSpecific from MSStorageDriver_ATAPISmartData where InstanceName=\"" + testDisk + "_0\"");
                // MSStorageDriver_FailurePredictData can be used in place of ATAPISmartData but the latter provides additional properties.
                ManagementObjectCollection FailDataSet = WMISearch.Get();
                foreach (ManagementObject attribute in FailDataSet)
                {
                    // We got what we needed, so save it, set the OK flag and break out of the foreach.
                    dataBuffer = (byte[])attribute.Properties["VendorSpecific"].Value;
                    ok = SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK;
                    break;
                }
            }
            catch
            {
                // WMI call failed; probably a "not found" or "not supported" which means the requested instance isn't in WMI so there's no data to fetch.
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_EXCEPTION;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }

        /// <summary>
        /// Gets the SMART thresholds for the specified disk from WMI, if WMI has that data available.
        /// </summary>
        /// <param name="physicalDrivePath">Path to the physical drive.</param>
        /// <param name="dataBuffer">The data buffer (probably will be 512-byte array) that is returned by the WMI call.</param>
        /// <returns>true if the SMART data was successfully retrieved from WMI; false otherwise.</returns>
        private static bool GetWmiSmartThresholds(String physicalDrivePath, out byte[] dataBuffer)
        {
            SMART_QUERY_RESULT ok = SMART_QUERY_RESULT.SMART_QUERY_PENDING;
            dataBuffer = null;

            try
            {
                ManagementObjectSearcher WMISearch = new ManagementObjectSearcher();
                WMISearch.Scope = new ManagementScope(@"\root\wmi");
                // Need to convert "\\" to "\\\\" or the WMI call will fail.
                String testDisk = physicalDrivePath.Replace("\\", "\\\\");

                // Not sure why but in WMI the instance name field has a _0 tacked on the end of it.
                WMISearch.Query = new ObjectQuery("select VendorSpecific from MSStorageDriver_FailurePredictThresholds where InstanceName=\"" + testDisk + "_0\"");
                ManagementObjectCollection FailDataSet = WMISearch.Get();
                foreach (ManagementObject attribute in FailDataSet)
                {
                    // We got what we needed, so save it, set the OK flag and break out of the foreach.
                    dataBuffer = (byte[])attribute.Properties["VendorSpecific"].Value;
                    ok = SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK;
                    break;
                }
            }
            catch
            {
                // WMI call failed; probably a "not found" or "not supported" which means the requested instance isn't in WMI so there's no data to fetch.
                dataBuffer = null;
                ok = SMART_QUERY_RESULT.SMART_QUERY_EXCEPTION;
            }
            finally
            {

            }
            return (ok == SMART_QUERY_RESULT.SMART_QUERY_RESULT_OK);
        }
    }
}
